from django.conf import settings
from django.urls import reverse_lazy as r_l
import os

root = '/easy_rest'
# if easy rest root specified in settings
if hasattr(settings, 'EASY_REST_ROOT_URL'):
    root = "/" + settings.EASY_REST_ROOT_URL


def reverse_lazy(pattern):
    app, name = pattern.split(":")
    if app == "easy_rest":
        return os.path.join(root, name)
    return r_l(pattern)



import sys
import json
import traceback
from django.urls import reverse_lazy
from django.views.generic import TemplateView
from ..utils.utils import reverse_lazy
from uuid import uuid4
from rest_framework.response import Response
from django.http.response import HttpResponseForbidden
from django.conf import settings


class DebugHandler(object):
    def __init__(self, request, data, status):
        self.request = request
        self.data = data
        self.status = status

    def handle(self):
        token = str(uuid4())
        self.request.session['debug_info_api'] = {
            "token": token,
            "data": self.data
        }
        # force save
        self.request.session.save()
        self.data['debug_url'] = reverse_lazy("easy_rest:debugger") + "?token={}".format(token)
        return Response(data=self.data, status=self.status)


class DebugObject(object):
    def __init__(self, tb, handler):
        self.tb = tb
        self.handler = handler
        exc_type, exc_value, exc_traceback = sys.exc_info()
        self.trace = traceback.format_exception(exc_type, exc_value, exc_traceback)

    def serialize(self):
        return {
            "error": repr(self.tb),
            "trace": self.get_trace(),
            "handler": self.handler,
        }

    def get_trace(self):
        return [t.lstrip().rstrip().replace("\n", "").replace("\r", "") for t in self.trace]


class DebugCache(object):
    def __init__(self):
        self.cache = []

    def update(self, tb, handler):
        self.cache.append(DebugObject(tb, handler).serialize())

    def serialize(self):
        data = []
        for obj in self.cache:
            data.append(obj)
        return data


class TbFile(object):
    data = ""

    def write(self, data):
        self.data += data

    def read(self):
        return self.data


def get_error(error):
    if hasattr(error, '__traceback__') and hasattr(error, 'tb_frame'):
        trace = ''.join(traceback.format_tb(error.__traceback__))
    else:
        trace = 'Unknown traceback object type({}) has no attribute __traceback__'.format(type(error))

    return trace


def create_trace(last_error):
    error = last_error['error']
    handler = last_error['handler']
    del last_error['error']
    del last_error['handler']

    return {
        "type": get_error(error),
        "details": traceback.format_exception(**last_error),
        "handler": handler
    }


class DebugView(TemplateView):
    template_name = "easy_rest/debug.html"

    def get(self, request, *args, **kwargs):
        debug_data = self.request.session.get("debug_info_api")
        if not debug_data:
            return HttpResponseForbidden()
        token = request.GET.get("token")

        if token != debug_data['token']:
            if settings.DEBUG:
                print("real token is ", debug_data['token'], "got", token)
            return HttpResponseForbidden("Invalid token")

        response = super(DebugView, self).get(request, *args, **kwargs)

        return response

    def get_context_data(self, **kwargs):
        debug_data = self.request.session.get("debug_info_api")['data']
        ctx = super(DebugView, self).get_context_data(**kwargs)
        if "input" in debug_data:
            request = debug_data["input"]
            request_data = request['request_data']
            ctx["api_input"] = json.dumps(request_data, indent=1)
            ctx["api_code"] = request["request_code"]
        if 'debug_url' in debug_data:
            del debug_data['debug_url']
        if "input" in debug_data:
            del debug_data['input']
        ctx['output'] = json.dumps(debug_data, indent=1)
        tb = self.request.session.get("last_debug_error")
        if tb:
            ctx['traceback'] = json.dumps(tb, indent=1)

        return ctx



import json


class DynamicEncoder(json.JSONEncoder):

    def default(self, obj):
        if hasattr(obj, "serialize"):
            return obj.serialize()

        return (
            "Can't serialize object, in order to serialize it "
            "Declare a serialize method in your class"
        )



from django import template
from django.conf import settings
from django.utils.safestring import mark_safe

register = template.Library()

# message to say on tag load
load_message = ''
# the static root of this app
root = '/easy_rest/static'

# if easy rest root specified in settings
if hasattr(settings, 'EASY_REST_ROOT_URL'):
    root = "/" + settings.EASY_REST_ROOT_URL + "/static"
else:
    load_message = '<!--{0}-->\n<script>console.warn("{0}")</script>'.format(
        "EASY_REST_ROOT_URL not specified in settings defaulting with easy_rest ")

# js base html
js_base = '<script src="' + root + '/{}"></script>'

# css base html
css_base = '<link rel="stylesheet" href="' + root + '/{}">'


@register.simple_tag(takes_context=True)
def load_rest_scripts(context):
    """
    loads only easy rest scripts
    :return: html with the scripts
    """
    data = _get_rest_scripts(context)
    if load_message:
        data = load_message + "\n" + data
    return mark_safe(data)


@register.simple_tag(takes_context=True)
def load_rest_all(context):
    """
    load rest scripts along with bootstrap
    :return: html including bootstrap and rest scripts
    """
    data = _get_rest_scripts(context) + _get_bootstrap()
    if load_message:
        data = load_message + "\n" + data
    return mark_safe(data)


def _load_debug_scripts():
    data = [js_base.format("jquery-3.2.1.min.js")]
    for script in [
        "bootstrap.js",
        "highlight.js",
        "json.js"
    ]:
        data.append(js_base.format("/highlight/{}".format(script)))
    for script in [
        "default.css",
        "bootstrap.css",
    ]:
        data.append(css_base.format("/highlight/{}".format(script)))
    return mark_safe("\n".join(data))


@register.simple_tag(takes_context=True)
def load_debug_scripts(context):
    return _load_debug_scripts() + load_rest_all(context)


def _get_rest_scripts(context=None):
    """
    :return: html scripts
    """
    files = []
    if settings.DEBUG:
        files.append("debugger.js")
    files += [
        'jquery-3.2.1.min.js',
        'restConsts.js',
        'Request.js',
        'PostHandler.js',
        'Submit.js',
        'fetch/restFetch.js'
    ]
    if context and context.get("js_context"):
        files.append("context.js")
    return '<!--start of easy rest scripts-->\n{}\n<!--end of easy rest scripts-->'.format(
        '\n'.join([js_base.format(file) for file in files])
    )


def _get_bootstrap():
    """
    :return: html scripts
    """
    files = [
        'bootstrap.css',
        'bootstrap-grid.css',
        'bootstrap-reboot.css',
    ]
    data = '\n'.join([css_base.format("bootstrap-4.0.0-alpha.6-dist/css/{}".format(file)) for file in files])
    data += '\n' + css_base.format("bootstrap-4.0.0-alpha.6-dist/js/{}".format("bootstrap.js"))
    return '<!--start bootstrap-->\n{}\n<!--end of bootstrap-->'.format(data)


@register.tag('livecontext')
def live_context(parser, token):
    bits = token.split_contents()[1:]  # first is the template tag name
    url = None
    if bits:
        url = bits[0]

    nodelist = parser.parse(('endlivecontext',))
    parser.delete_first_token()
    return LiveContext(nodelist, url)


class LiveContext(template.Node):
    def __init__(self, nodelist, url):
        self.nodelist = nodelist
        self.url = url

    def render(self, context):
        output = self.nodelist.render(context)
        if self.url:
            return '<div class="fetch-context" data-fetch-url="{}">\n{}\n</div>'.format(self.url, output)
        else:
            return '<div class="fetch-context">\n{}\n</div>'.format(output)
